/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001  Ludvig Strigeus
 * Copyright (C) 2001-2004 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /cvsroot/scummvm/scummvm/backends/sdl/sdl.cpp,v 1.56.2.2 2004/02/28 13:01:43 fingolfin Exp $
 *
 */

#include "sdl-common.h"
#include "common/scaler.h"
#include "common/util.h"

class OSystem_SDL : public OSystem_SDL_Common {
public:
	OSystem_SDL();

	// Update the dirty areas of the screen
	void update_screen();

	// Set a parameter
	uint32 property(int param, Property *value);

protected:
	SDL_Surface *_hwscreen;    // hardware screen

	ScalerProc *_scaler_proc;

	virtual void load_gfx_mode();
	virtual void unload_gfx_mode();
	virtual bool save_screenshot(const char *filename);
	void hotswap_gfx_mode();
};

OSystem_SDL_Common *OSystem_SDL_Common::create_intern() {
	return new OSystem_SDL();
}

OSystem_SDL::OSystem_SDL()
	 : _hwscreen(0), _scaler_proc(0)
{
}
#include <Mda\Common\Audio.h>
#include <MdaAudioOutputStream.h>
class OSystem_SDL_Symbian:public OSystem_SDL,public MMdaAudioOutputStreamCallback
{
public:
	OSystem_SDL_Symbian();
	~OSystem_SDL_Symbian();
	static OSystem *create(int gfx_mode);
	RThread* gSndThread;
	static TInt Sound_and_music_thread(TAny *params);
	void  sound_and_music_thread(TAny *params);

	void MaoscOpenComplete(TInt aError) ;
	void MaoscBufferCopied(TInt aError, const TDesC8& aBuffer);
	void MaoscPlayComplete(TInt aError);
	// Set function that generates samples 
	bool set_sound_proc( SoundProc proc, void *param, SoundFormat format);
	void clear_sound_proc();

	CMdaAudioOutputStream* iStream;
	CActiveScheduler* iScheduler ;
	TBool iTerminateSoundThread;
	TBool iSchedStarted;
	TBool FM_high_quality;
protected:
	static OSystem_SDL_Symbian *create_intern();

};

void OSystem_SDL::load_gfx_mode() {
	_forceFull = true;
	_mode_flags |= DF_UPDATE_EXPAND_1_PIXEL;

	_tmpscreen = NULL;
	_tmpScreenWidth = (_screenWidth + 3);
	
	switch(_mode) {
	case GFX_NORMAL:
		_scaleFactor = 1;
		_scaler_proc = Normal1x;
		break;
		/*
	case GFX_DOUBLESIZE:
		_scaleFactor = 2;
		_scaler_proc = Normal2x;
		break;
	case GFX_TRIPLESIZE:
		_scaleFactor = 3;
		_scaler_proc = Normal3x;
		break;

	case GFX_2XSAI:
		_scaleFactor = 2;
		_scaler_proc = _2xSaI;
		break;
	case GFX_SUPER2XSAI:
		_scaleFactor = 2;
		_scaler_proc = Super2xSaI;
		break;
	case GFX_SUPEREAGLE:
		_scaleFactor = 2;
		_scaler_proc = SuperEagle;
		break;
	case GFX_ADVMAME2X:
		_scaleFactor = 2;
		_scaler_proc = AdvMame2x;
		break;
	case GFX_ADVMAME3X:
		_scaleFactor = 3;
		_scaler_proc = AdvMame3x;
		break;
	case GFX_HQ2X:
		_scaleFactor = 2;
		_scaler_proc = HQ2x;
		break;
	case GFX_HQ3X:
		_scaleFactor = 3;
		_scaler_proc = HQ3x;
		break;
	case GFX_TV2X:
		_scaleFactor = 2;
		_scaler_proc = TV2x;
		break;
	case GFX_DOTMATRIX:
		_scaleFactor = 2;
		_scaler_proc = DotMatrix;
		break;
*/
	default:
		error("unknown gfx mode %d", _mode);
	}

	//
	// Create the surface that contains the 8 bit game data
	//
	_screen = SDL_CreateRGBSurface(SDL_SWSURFACE, _screenWidth, _screenHeight, 8, 0, 0, 0, 0);
	if (_screen == NULL)
		error("_screen failed");

	//
	// Create the surface that contains the scaled graphics in 16 bit mode
	//

	_hwscreen = SDL_SetVideoMode(_screenWidth * _scaleFactor, (_adjustAspectRatio ? 240 : _screenHeight) * _scaleFactor, 16, 
		_full_screen ? (SDL_FULLSCREEN|SDL_SWSURFACE) : SDL_SWSURFACE
	);
	if (_hwscreen == NULL) {
		// DON'T use error(), as this tries to bring up the debug
		// console, which WON'T WORK now that _hwscreen is hosed.

		// FIXME: We should be able to continue the game without
		// shutting down or bringing up the debug console, but at
		// this point we've already screwed up all our member vars.
		// We need to find a way to call SDL_VideoModeOK *before*
		// that happens and revert to all the old settings if we
		// can't pull off the switch to the new settings.
		//
		// Fingolfin says: the "easy" way to do that is not to modify
		// the member vars before we are sure everything is fine. Think
		// of "transactions, commit, rollback" style... we use local vars
		// in place of the member vars, do everything etc. etc.. In case
		// of a failure, rollback is trivial. Only if everything worked fine
		// do we "commit" the changed values to the member vars.
		warning("SDL_SetVideoMode says we can't switch to that mode");
		quit();
	}

	//
	// Create the surface used for the graphics in 16 bit before scaling, and also the overlay
	//

	// Distinguish 555 and 565 mode
	if (_hwscreen->format->Rmask == 0x7C00)
		InitScalers(555);
	else
		InitScalers(565);
	
	// Need some extra bytes around when using 2xSaI
	uint16 *tmp_screen = (uint16 *)calloc(_tmpScreenWidth * (_screenHeight + 3), sizeof(uint16));
	_tmpscreen = SDL_CreateRGBSurfaceFrom(tmp_screen,
						_tmpScreenWidth, _screenHeight + 3, 16, _tmpScreenWidth * 2,
						_hwscreen->format->Rmask,
						_hwscreen->format->Gmask,
						_hwscreen->format->Bmask,
						_hwscreen->format->Amask);

	if (_tmpscreen == NULL)
		error("_tmpscreen failed");

	// keyboard cursor control, some other better place for it?
	km.x_max = _screenWidth * _scaleFactor - 1;
	km.y_max = (_adjustAspectRatio ? 240 : _screenHeight) * _scaleFactor - 1;
	km.delay_time = 25;
	km.last_time = 0;
}

void OSystem_SDL::unload_gfx_mode() {
	if (_screen) {
		SDL_FreeSurface(_screen);
		_screen = NULL; 
	}

	if (_hwscreen) {
		SDL_FreeSurface(_hwscreen); 
		_hwscreen = NULL;
	}

	if (_tmpscreen) {
		free(_tmpscreen->pixels);
		SDL_FreeSurface(_tmpscreen);
		_tmpscreen = NULL;
	}
}

void OSystem_SDL::hotswap_gfx_mode() {
	if (!_screen)
		return;

	// Keep around the old _screen & _tmpscreen so we can restore the screen data
	// after the mode switch.
	SDL_Surface *old_screen = _screen;
	SDL_Surface *old_tmpscreen = _tmpscreen;

	// Release the HW screen surface
	SDL_FreeSurface(_hwscreen); 

	// Setup the new GFX mode
	load_gfx_mode();

	// reset palette
	SDL_SetColors(_screen, _currentPalette, 0, 256);

	// Restore old screen content
	SDL_BlitSurface(old_screen, NULL, _screen, NULL);
	SDL_BlitSurface(old_tmpscreen, NULL, _tmpscreen, NULL);
	
	// Free the old surfaces
	SDL_FreeSurface(old_screen);
	free(old_tmpscreen->pixels);
	SDL_FreeSurface(old_tmpscreen);

	// Blit everything to the screen
	update_screen();
	
	// Make sure that an EVENT_SCREEN_CHANGED gets sent later
	_modeChanged = true;
}

void OSystem_SDL::update_screen() {
	assert(_hwscreen != NULL);

	Common::StackLock lock(_graphicsMutex, this);	// Lock the mutex until this function ends

	// If the shake position changed, fill the dirty area with blackness
	if (_currentShakePos != _newShakePos) {
		SDL_Rect blackrect = {0, 0, _screenWidth * _scaleFactor, _newShakePos * _scaleFactor};

		if (_adjustAspectRatio)
			blackrect.h = real2Aspect(blackrect.h - 1) + 1;

		SDL_FillRect(_hwscreen, &blackrect, 0);

		_currentShakePos = _newShakePos;

		_forceFull = true;
	}

	// Make sure the mouse is drawn, if it should be drawn.
	draw_mouse();
	
	// Check whether the palette was changed in the meantime and update the
	// screen surface accordingly. 
	if (_paletteDirtyEnd != 0) {
		SDL_SetColors(_screen, _currentPalette + _paletteDirtyStart, 
			_paletteDirtyStart,
			_paletteDirtyEnd - _paletteDirtyStart);
		
		_paletteDirtyEnd = 0;

		_forceFull = true;
	}

	// Force a full redraw if requested
	if (_forceFull) {
		_num_dirty_rects = 1;

		_dirty_rect_list[0].x = 0;
		_dirty_rect_list[0].y = 0;
		_dirty_rect_list[0].w = _screenWidth;
		_dirty_rect_list[0].h = _screenHeight;
	}

	// Only draw anything if necessary
	if (_num_dirty_rects > 0) {

		SDL_Rect *r; 
		SDL_Rect dst;
		uint32 srcPitch, dstPitch;
		SDL_Rect *last_rect = _dirty_rect_list + _num_dirty_rects;

		if (_scaler_proc == Normal1x && !_adjustAspectRatio) {
			SDL_Surface *target = _overlayVisible ? _tmpscreen : _screen;
			for (r = _dirty_rect_list; r != last_rect; ++r) {
				dst = *r;
				
				if (_overlayVisible) {
					// FIXME: I don't understand why this is necessary...
					dst.x--;
					dst.y--;
				}
				dst.y += _currentShakePos;
				if (SDL_BlitSurface(target, r, _hwscreen, &dst) != 0)
					error("SDL_BlitSurface failed: %s", SDL_GetError());
			}
		} else {
			if (!_overlayVisible) {
				for (r = _dirty_rect_list; r != last_rect; ++r) {
					dst = *r;
					dst.x++;	// Shift rect by one since 2xSai needs to acces the data around
					dst.y++;	// any pixel to scale it, and we want to avoid mem access crashes.
					if (SDL_BlitSurface(_screen, r, _tmpscreen, &dst) != 0)
						error("SDL_BlitSurface failed: %s", SDL_GetError());
				}
			}

			SDL_LockSurface(_tmpscreen);
			SDL_LockSurface(_hwscreen);

			srcPitch = _tmpscreen->pitch;
			dstPitch = _hwscreen->pitch;

			for (r = _dirty_rect_list; r != last_rect; ++r) {
				register int dst_y = r->y + _currentShakePos;
				register int dst_h = 0;
				register int orig_dst_y = 0;

				if (dst_y < _screenHeight) {
					dst_h = r->h;
					if (dst_h > _screenHeight - dst_y)
						dst_h = _screenHeight - dst_y;

					dst_y *= _scaleFactor;

					if (_adjustAspectRatio) {
						orig_dst_y = dst_y;
						dst_y = real2Aspect(dst_y);
					}

					_scaler_proc((byte *)_tmpscreen->pixels + (r->x * 2 + 2) + (r->y + 1) * srcPitch, srcPitch,
						(byte *)_hwscreen->pixels + r->x * 2 * _scaleFactor + dst_y * dstPitch, dstPitch, r->w, dst_h);
				}

				r->x *= _scaleFactor;
				r->y = dst_y;
				r->w *= _scaleFactor;
				r->h = dst_h * _scaleFactor;

				//if (_adjustAspectRatio && orig_dst_y / _scaleFactor < _screenHeight)
				//	r->h = stretch200To240((uint8 *) _hwscreen->pixels, dstPitch, r->w, r->h, r->x, r->y, orig_dst_y);
			}
			SDL_UnlockSurface(_tmpscreen);
			SDL_UnlockSurface(_hwscreen);
		}

		// Readjust the dirty rect list in case we are doing a full update.
		// This is necessary if shaking is active.
		if (_forceFull) {
			_dirty_rect_list[0].y = 0;
			_dirty_rect_list[0].h = (_adjustAspectRatio ? 240 : _screenHeight) * _scaleFactor;
		}

		// Finally, blit all our changes to the screen
		SDL_UpdateRects(_hwscreen, _num_dirty_rects, _dirty_rect_list);
	}

	_num_dirty_rects = 0;
	_forceFull = false;
}

uint32 OSystem_SDL::property(int param, Property *value) {

	Common::StackLock lock(_graphicsMutex, this);	// Lock the mutex until this function ends

	if (param == PROP_TOGGLE_FULLSCREEN) {
		assert(_hwscreen != 0);
		_full_screen ^= true;

		if (_mouseDrawn)
			undraw_mouse();

#ifdef MACOSX
		// On OS X, SDL_WM_ToggleFullScreen is currently not implemented. Worse,
		// it still always returns -1. So we simply don't call it at all and
		// use hotswap_gfx_mode() directly to switch to fullscreen mode.
		hotswap_gfx_mode();
#else
		if (!SDL_WM_ToggleFullScreen(_hwscreen)) {
			// if ToggleFullScreen fails, achieve the same effect with hotswap gfx mode
			hotswap_gfx_mode();
		}
#endif
		return 1;
	} else if (param == PROP_SET_GFX_MODE) {
		if (value->gfx_mode > 11)	// FIXME! HACK, hard coded threshold, not good
			return 0;

		_mode = value->gfx_mode;
		hotswap_gfx_mode();

		return 1;
	} else if (param == PROP_TOGGLE_ASPECT_RATIO) {
		if (_screenHeight == 200) {
			assert(_hwscreen != 0);
			_adjustAspectRatio ^= true;
			hotswap_gfx_mode();
		}
	} else if (param == PROP_HAS_SCALER) {
		if (value->gfx_mode <= 11)	// FIXME: Hardcoded
			return 1;
		return 0;
	}

	return OSystem_SDL_Common::property(param, value);
}

bool OSystem_SDL::save_screenshot(const char *filename) {
	assert(_hwscreen != NULL);

	Common::StackLock lock(_graphicsMutex, this);	// Lock the mutex until this function ends
	SDL_SaveBMP(_hwscreen, filename);
	return true;
}



#include "sdl-common.h"



#define FRAG_SIZE 4092
typedef struct {
	OSystem::SoundProc sound_proc;
	void *param;
	byte format;
	OSystem_SDL_Symbian* iSystem;
} THREAD_PARAM;

OSystem *OSystem_SDL_Symbian::create(int gfx_mode) 
{
	OSystem_SDL_Symbian *syst = OSystem_SDL_Symbian::create_intern();
	syst->init_intern(gfx_mode);

	return syst;
}

OSystem_SDL_Symbian *OSystem_SDL_Symbian::create_intern() {
	return new OSystem_SDL_Symbian();
}
#include "config-manager.h"

extern Common::ConfigManager *g_config;

OSystem_SDL_Symbian::OSystem_SDL_Symbian()
{
	const char *dir = NULL;
/*	dir = g_config->get("savepath");

	// If SCUMMVM_SAVEPATH was not specified, try to use general path from config
	if (!dir || dir[0] == 0)
		dir = g_config->get("savepath", "scummvm");
	if (!dir || dir[0] == 0)
	{
		setenv("SCUMMVM_SAVEPATH","C:\\documents\\",true);
		setenv("savepath","C:\\documents\\",true);
	}
	else
	{
		setenv("SCUMMVM_SAVEPATH",dir,true);
		setenv("savepath",dir,true);
	}*/
#pragma message("Set save path here")
}

OSystem_SDL_Symbian::~OSystem_SDL_Symbian()
{
}


void  OSystem_SDL_Symbian::sound_and_music_thread(TAny *params)
{
	iTerminateSoundThread=EFalse;
	iScheduler = new (ELeave) CActiveScheduler;
	CActiveScheduler::Install(iScheduler);

	iStream = CMdaAudioOutputStream::NewL(*this);
	TMdaAudioDataSettings audioSettings;
	audioSettings.Query();
#ifdef SYMBIAN_SERIES_60
	audioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate8000Hz;
	audioSettings.iChannels = TMdaAudioDataSettings::EChannelsMono; // only support mono in s60
#else
	audioSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate22050Hz;
	audioSettings.iChannels = TMdaAudioDataSettings::EChannelsStereo;
#endif
	audioSettings.iFlags = 0;
	audioSettings.iVolume = 0;

	iStream->Open(&audioSettings);
	iSchedStarted=ETrue;
	CActiveScheduler::Start();

	int  param, frag_size;
	uint8* sound_buffer=(uint8*)malloc(FRAG_SIZE);
	OSystem::SoundProc sound_proc = ((THREAD_PARAM *) params)->sound_proc;
	void *proc_param = ((THREAD_PARAM *) params)->param;
	while (1) {
		uint8 *buf = (uint8 *)sound_buffer;
		int size, written;
		sound_proc(proc_param, (byte *)sound_buffer, FRAG_SIZE);
		size = FRAG_SIZE;
#ifdef SYMBIAN_SERIES_60
		// series 60 only supports mono for now so we fix the sample
		TUint32 sample = 0;
		TUint16 *srcBuf = (TUint16 *) sound_buffer;
		TUint16 *dstBuf = (TUint16 *) sound_buffer;

		for(TInt i = 0; i < FRAG_SIZE / 2; i += 6)
		{
			sample = *(srcBuf + i);
			sample += *(srcBuf + i + 1);
			*dstBuf = sample / 2; // average of six samples
			dstBuf++;
		}

		size = FRAG_SIZE / 6;
#endif
		
		TPtrC8 ptr (buf,size);
		
		iStream->WriteL(ptr);//write(sound_fd, buf, size);
		iSchedStarted=ETrue;
		CActiveScheduler::Start();
		
		if(iTerminateSoundThread)
		{
			break;
		}
	}

	iStream->Stop();
	delete iStream;
	iStream = NULL;
	delete iScheduler;
	iScheduler = NULL;
	free(sound_buffer);
	iTerminateSoundThread =EFalse;
}

TInt OSystem_SDL_Symbian::Sound_and_music_thread(TAny *params)
{
	CTrapCleanup* cleanup = CTrapCleanup::New();
	TRAPD(err,((THREAD_PARAM *) params)->iSystem->sound_and_music_thread(params));
	delete cleanup;

	return 0;
}

bool OSystem_SDL_Symbian::set_sound_proc(SoundProc proc, void *param, SoundFormat format)
{
	static THREAD_PARAM thread_param;

	/* And finally start the music thread */
	thread_param.param = param;
	thread_param.sound_proc = proc;
	thread_param.format = format;
	thread_param.iSystem=this;

	TInt minStack = 8192;

	gSndThread = new RThread;
	gSndThread->Create(_L("SCUSND"),Sound_and_music_thread,minStack,NULL,(void *)&thread_param);
	gSndThread->SetPriority(EPriorityMore);
	gSndThread->Resume();
	return true;
}

void OSystem_SDL_Symbian::MaoscOpenComplete(TInt /*aError*/) 
{
#ifdef SYMBIAN_SERIES_60
	iStream->SetAudioPropertiesL( TMdaAudioDataSettings::ESampleRate8000Hz,TMdaAudioDataSettings::EChannelsMono); // Set 11025 stereo sound, 16 bit only supported
#else
	iStream->SetAudioPropertiesL( TMdaAudioDataSettings::ESampleRate22050Hz,TMdaAudioDataSettings::EChannelsStereo); // Set 11025 stereo sound, 16 bit only supported
#endif
	iStream->SetVolume(iStream->MaxVolume());
	if(iSchedStarted)
	{
	CActiveScheduler::Stop();
	iSchedStarted=EFalse;
	}
}

void OSystem_SDL_Symbian::MaoscBufferCopied(TInt /*aError*/, const TDesC8& /*aBuffer*/)
{
	if(iSchedStarted)
	{
	CActiveScheduler::Stop();
	iSchedStarted=EFalse;
	}
}

void OSystem_SDL_Symbian::MaoscPlayComplete(TInt /*aError*/)
{
	if(iSchedStarted)
	{
	CActiveScheduler::Stop();
	iSchedStarted=EFalse;
	}
}

void OSystem_SDL_Symbian::clear_sound_proc()
{
	if(gSndThread)
	{
		iTerminateSoundThread =ETrue;
		while(iTerminateSoundThread)
		{
			User::After(10000);
		}
		gSndThread->Terminate(0);
		gSndThread->Close();
		delete gSndThread;
		gSndThread  = NULL;
	}
}

OSystem *OSystem_SDL_create(int gfx_mode) {
	#ifdef __SYMBIAN32__
	return OSystem_SDL_Symbian::create(gfx_mode);
	#else
	return OSystem_SDL_Common::create(gfx_mode);
	#endif
}
